import socket
import requests
# import json
import psutil

from .html_templates import * # pylint: disable=unused-wildcard-import



STYLES = '''
<style>
	body {
		width:100%;
		height:100%;
		margin: 0;
		display:flex;
		flex-direction:column;
		align-items:center;
	}
	.header-bar {
		display:flex;
		justify-content: center;
		align-items:center;
		position: sticky;
		top: 0px;
		width:100%;
		margin-bottom: 10px;
	}
	.wrapper {
		margin: 50px;
		display: grid;
		grid-template-columns: repeat(3, 1fr);
		gap: 30px;
	}
	.monitor-block {
		width: 300px;
		height: 120px;
		border: thin solid lightgrey;
		border-radius: 5px;
		box-shadow: 2px 2px 4px #888888;
		display:flex;
		flex-direction:column;
		justify-content: center;
		align-items:center;
		cursor:pointer;
	}
	.monitor-block:hover {
		box-shadow: 5px 5px 5px #888888;
	}
	.error-border {
		border: 2px solid #f53b3b;
	}
	.block-title {
		flex: 2;
		display:flex;
		justify-content: center;
		align-items:center;
	}
	.block-msg {
		flex: 1;
		display:flex;
		flex-direction:row;
		justify-content: flex-end;
		align-items: center;
		width:100%;
		background-color: #eee;
	}
	.block-msg > div {
		display:flex;
		justify-content: center;
		align-items: center;
		padding: 2px;
		margin: 5px;
		width:25px;
		height:25px;
		background-color: white;
		border-radius: 50%;
	}
	.error-msg {
		background-color: #f53b3b !important;
		border: none !important;
		color: white;
	}
</style>
'''


class ControlPanel:
	'''Automatically scan ports and consolidate many TaskMonitors'''

	def __init__(self,
		app,
		ports=[],
		external_addrs=[],
		page_refresh=60):

		self.app = app
		self.machine = socket.gethostname()
		self.local_ip = socket.gethostbyname(self.machine)
		self.ports = set(ports)
		self.external_addrs = set(external_addrs)
		self.page_refresh = page_refresh
		self.app.add_url_rule("/", view_func=self._render_monitors, methods=['GET'])


	def scan(self, min_port=1000, max_port=10000, timeout=5):
		for conn in psutil.net_connections():
			if conn.status == "LISTEN" and conn.laddr.port >= min_port and conn.laddr.port <= max_port:
				m = self._get_taskmonitor(self.local_ip, conn.laddr.port, timeout=timeout)
				if m is not None:
					self.ports.add(conn.laddr.port)
				print('>> scanned', conn.laddr.port, "- found" if m is not None else "")


	def _get_taskmonitor(self, host, port, timeout=5):
		try:
			monitor_url = f"http://{host}:{port}/@taskmonitor" # need to add option to change this endpoint since task monitor has that option
			res = requests.get(f"{monitor_url}/json/summary", timeout=timeout).json()
			# print(json.dumps(res, indent=4))
			res['port'] = port
			res['url'] = monitor_url
			return res
		except Exception:
			# print(e)
			return None


	def _iter_monitors(self):
		for port in self.ports:
			monitor = self._get_taskmonitor(self.local_ip, port)
			if monitor is not None:
				yield monitor
		for host, port in self.external_addrs:
			monitor = self._get_taskmonitor(host, port)
			if monitor is not None:
				yield monitor


	def _render_monitors(self):
		content = []
		tot_jobs = 0
		for monitor in self._iter_monitors():
			css = ['monitor-block']
			attrs = {}
			elem = ""
			if 'error' in monitor:
				elem = H(5, monitor['error']) + SPAN(str(monitor['url']))
				css.append('error-border')
				css.append('no-page')
				attrs['title'] = monitor['error']
			else:
				mon = monitor['success']
				err_msg_css = []
				if mon['summary']['errors'] > 0:
					css.append('error-border')
					err_msg_css.append('error-msg')
				msg = f"tasks: {DIV(mon['summary']['count'])}  errors: {DIV(mon['summary']['errors'], css=err_msg_css)}"
				elem = SPAN(B(mon['name']), css=['block-title']) + SPAN(msg, css=['block-msg'])
				attrs['data-url'] = monitor['url']
				attrs['title'] = f"{mon['name']}\n{monitor['url']}"
			content.append(DIV(elem, css=css, attrs=attrs))
			tot_jobs += mon['summary']['count']
		wrapper = DIV(''.join(content), css='wrapper')
		header_txt = f"Control Panel"
		header = DIV(H(2, header_txt), css=['header-bar'])
		summary_txt = SMALL(f"Monitoring {tot_jobs} jobs")
		rerun_txt = SMALL(f"Auto-refresh in {SPAN(self.page_refresh, attrs={'id': 'refresh-msg'})} seconds")

		auto_reload = SCRIPT('''
		let COUNT_DOWN = {page_refresh}
		window.addEventListener('load', (event) => {{
			const timer = setInterval(()=>{{
				if (COUNT_DOWN > 0) {{
					COUNT_DOWN --
					document.getElementById('refresh-msg').innerText = COUNT_DOWN
				}} else {{
					clearInterval(timer)
					location.reload()
				}}
			}}, 1000)
			document.querySelectorAll('.monitor-block:not(.no-page)').forEach(block=>{{
				block.addEventListener('click', ()=>{{
					window.location.href=block.getAttribute("data-url")
				}})
			}})
		}});
		'''.format(page_refresh=self.page_refresh))
		return HTML(''.join([
			STYLES,
			header,
			summary_txt,
			rerun_txt,
			wrapper,
			auto_reload
		]), title=header_txt)



from datetime import datetime as dt
from collections import OrderedDict
import json
import random
import string

from dateutil import tz
from flask import request


from .html_templates import * # pylint: disable=unused-wildcard-import


STYLES = '''
<style>
	html {
		--console-bg: #333333;
		--thumb-bg: grey;
	}
	body {
		width:100%;
		height:100%;
		margin: 0;
		display:flex;
		flex-direction:column;
		align-items:center;
		font-family: sans-serif;
		font-size: 12px;
	}
	*::-webkit-scrollbar {
		width: 14px !important;
		height: 14px !important;
	}
	* {
		scrollbar-width: thin;
		scrollbar-color: var(--thumb-bg) var(--console-bg);
	}
	*::-webkit-scrollbar-track {
		background: var(--console-bg);
	}
	*::-webkit-scrollbar-thumb {
		background-color: var(--thumb-bg) ;
		border-radius: 6px;
		border: 3px solid var(--console-bg);
	}
	*::-webkit-scrollbar-corner {
		background-color: var(--console-bg) ;
	}
	th[role=columnheader]:not(.no-sort) { /*tablesort.css*/
		cursor: pointer;
	}
	th[role=columnheader]:not(.no-sort):after { /*tablesort.css*/
		content: '';
		float: right;
		margin-top: 7px;
		border-width: 4px 4px 0;
		border-style: solid;
		border-color: #e3e3e3 transparent;
		visibility: hidden;
		-ms-user-select: none;
		-webkit-user-select: none;
		-moz-user-select: none;
		user-select: none;
	}
	th[aria-sort=ascending]:not(.no-sort):after { /*tablesort.css*/
		border-width: 0 4px 4px;
	}
	th[aria-sort]:not(.no-sort):after { /*tablesort.css*/
		visibility: visible;
	}
	small, h2 { padding-top:10px;}
	table {
		border-spacing: 5px;
		border-collapse: collapse;
		border: 1px solid #c2c2c2;
		width:85%;
		margin-top:20px;
		margin-bottom:40px;
	}
	tr:hover { background-color: #ededed; }
	td, th {
		border: 1px solid #d1d1d1;
		padding: 5px;
	}
	th {
		height:2vh;
		background-color:var(--console-bg);
		color:white;
	}
	td.grey {color:#c2c2c2;}
	td.yellow {background-color:yellow;}
	td.green {
		background-color:#d2ffcc;
		color:green;
	}
	td.red {
		background-color:red;
		color:white;
		font-weight:bold;
	}
	td.blue {
		background-color:blue;
		color:white;
		font-weight:bold;
	}
	tr.row-hidden { display:none; }
	a > button {
		width:100%;
		height:100%;
		cursor:pointer;
	}
	.container {
		width:100%;
		height:100%;
		display:flex;
		flex-direction:row;
	}
	.center {
		justify-content:center;
		align-items:center;
	}
	.btn {
		cursor: pointer;
	}
	.btn:disabled,.btn[disabled] {
		cursor: not-allowed;
	}
	.monitor {
		width: 30vw;
		height: 100vh;
		display:flex;
		flex-flow: column;
		align-items:center;
	}
	.monitor > div {
		flex: 1 1 auto;
		width:100%;
		overflow-wrap: break-word;
	}
	.info_table {
		border:none;
		margin-bottom:30px;
		width:100%;
	}
	.info_table td {
		border:none;
		width:50%;
		text-align:left;
	}
	td.title {
		font-weight:bold;
		text-align:right !important;
		padding-right:20px;
	}
	.logs_div {
		width: 70vw;
		height: 100vh;
		display:flex;
		align-items:center;
		justify-content:center;
	}
	.log_table {
		table-layout:fixed;
		width:100%;
		height:100%;
		margin-top:0px;
		margin-bottom:0px;
		overflow:hidden;
	}
	.log_table td, .log_table th {
		border:none;
		border-left: 1px solid grey;
		vertical-align: top;
		overflow:hidden;
	}
	.console-div {
		width:100%;
		top:0px;
		bottom:0px;
		height:95vh;
		overflow:scroll;
		white-space: nowrap;
		list-style-type: none;
		font-size: 13px;
		padding-left: 5px;
	}
	.console-color {
		background-color:var(--console-bg);
		color:white;
	}
	.brdr {border: 1px solid grey;}
	pre, code {
		background-color:transparent !important;
		overflow:visible !important;
	}
	input[type=text] {
		width: 250px;
		padding: 6px 15px;
		margin: 8px 0;
		border: 1px solid #999;
		border-radius: 3px;
		outline: none;
	}
</style>
'''

class TaskMonitor(object):
	'''
	Web interface to monitor and manage tasks

	Args:
	- app (int): ``Flask`` application
	- sched (TaskScheduler): task scheduler with task definitions
	- display_name (str): name of the application to be displayed
		- default app.name

	- endpoint (str): URL endpoint where the taskmonitor can be viewed
		- default "@taskmonitor"
	- homepage_refresh (int): home page auto refresh interval (in seconds)
		- default 30
	- taskpage_refresh (int): task page auto refresh interval (in seconds)
		- default 5
	- can_rerun (bool): if True adds `rerun` button to job page
		- default True
	- can_disable (bool): if True adds `disable` button to job page
		- default True
	'''
	def __init__(self, app, sched, display_name=None, endpoint="@taskmonitor", homepage_refresh=30, taskpage_refresh=5,
		can_rerun=True, # adds rerun button to job page
		can_disable=True, # adds disable button to job page
		):
		self.tzname = sched._tz_default
		self._init_dt = dt.now(tz.gettz(self.tzname)).strftime("%m/%d/%Y %I:%M %p %Z") # preformatted start time
		self.app = app
		self.sched = sched
		self._endpoint = endpoint
		self._display_name = display_name or self.app.name
		self._homepage_refresh = homepage_refresh
		self._taskpage_refresh = taskpage_refresh
		self._api_protection_token = ''.join(random.choices(string.ascii_letters, k=20)) # protect job reruns external API calls. makes sure endpoint can only be called by TaskMonitor

		self._can_rerun = can_rerun
		self._can_disable = can_disable

		self.app.add_url_rule("/{}".format(self._endpoint), view_func=self.__show_all, methods=['GET'])
		self.app.add_url_rule("/{}/<int:n>".format(self._endpoint), view_func=self.__show_one, methods=['GET'])
		self.app.add_url_rule("/{}/rerun".format(self._endpoint), view_func=self.__rerun_job, methods=['POST'])
		self.app.add_url_rule("/{}/enable_disable".format(self._endpoint), view_func=self.__enable_disable_job, methods=['POST'])
		self.app.add_url_rule("/{}/json/all".format(self._endpoint), view_func=self.__get_all_json, methods=['GET'])
		self.app.add_url_rule("/{}/json/summary".format(self._endpoint), view_func=self.__get_summary_json, methods=['GET'])
		self.app.add_url_rule("/{}/json/<int:n>".format(self._endpoint), view_func=self.__get_one_json, methods=['GET'])

	def __html_wrap(self, *args):
		return HTML(''.join(args), title="{} Task Monitor".format(self._display_name))

	def __state(self, jdict):
		state = {'state':'READY', 'css': 'grey', 'title': '' }
		if jdict['is_disabled']:
			state['state'] = "DISABLED"
			state['css'] = "blue"
			return state # no need to check status if disabled

		if jdict['is_running']:
			state['state'] = "RUNNING"
			state['css'] = "yellow"
		elif jdict['logs']['err'].strip()!='':
			state['state'] = "ERROR"
			state['css'] = "red"
			state['title'] = jdict['logs']['err'].strip().split("\n")[-1]
		elif jdict['logs']['end'] is not None and jdict['logs']['log'].strip()!='':
			state['state'] = "SUCCESS"
			state['css'] = "green"
		return state

	def __duration(self, jdict):
		duration = None
		if jdict['logs']['start'] is not None and jdict['logs']['end'] is not None:
			seconds = (jdict['logs']['end']-jdict['logs']['start']).seconds
			if seconds >= 60:
				minutes = seconds // 60
				seconds = seconds % 60
				duration = "{}:{} minutes".format(minutes, str(seconds).zfill(2))
			elif seconds==1:
				duration = "{} second".format(seconds)
			else:
				duration = "{} seconds".format(seconds)
		return duration

	def __timestr_to_12hr(self, tstr):
		d = dt.strptime(dt.now().strftime("%Y-%m-%d "+ str(tstr)), "%Y-%m-%d %H:%M")
		return d.strftime("%I:%M%p")

	def __schedule_str(self, jdict):
		if isinstance(jdict['every'], int): # jdict['type']=='RepeatJob'
			out = "every {} seconds".format(jdict['every'])
		elif jdict['type']=='OneTimeJob':
			out = "on {} at {}".format(jdict['every'], jdict['at'])
		elif jdict['type']=='NeverJob':
			out = 'on-demand'
		elif isinstance(jdict['at'], (list,set,tuple)):
			out = "every {} at {}".format(jdict['every'], ', '.join(jdict['at']))
		else:
			out = "every {} at {}".format(jdict['every'], jdict['at'])

		if isinstance(jdict['tzname'], str):
			out += " " + dt.now(tz.gettz(jdict['tzname'])).strftime("[%Z]")
		return out

	def __date_fmt(self, d):
		fallback = '-'+('&nbsp;'*30) # a hiphen and some html spaces
		return d.strftime("%Y-%m-%d %H:%M:%S %Z") if d is not None else fallback

	def __date_sort_attr(self, d):
		return {'data-sort': d.timestamp() if d is not None else 0}

	def __descrTD(self, d):
		if d is None: return TD('-')
		d = d.strip()
		short_d = d[:30] + "..."
		return '<td title="{}">{}</td>'.format(
			d,
			short_d if len(d)>30 else d
		)

	def __src_err_line(self, j):
		if j['logs']['err'].strip()!='':
			for l in j['logs']['err'].strip().split("\n"):
				if l.strip():
					idx = j['src'].find(l.strip())
					if idx>=0:
						return j['src'][:idx].count("\n")
		return -1

	def __get_all_json(self):
		if len(self.sched.jobs)==0:
			return json.dumps({'error':'Nothing here'})
		else:
			return json.dumps({'success': [j.to_dict() for j in self.sched.jobs]}, default=str)

	def __get_summary_json(self):
		if len(self.sched.jobs)==0:
			return json.dumps({'error':'Nothing here'})
		else:
			details = []
			summary = {'count': 0, 'errors': 0}
			for j in self.sched.jobs:
				jd = j.to_dict()
				state = self.__state(jd)
				summary['count'] += 1
				if state['state'] == "ERROR":
					summary['errors'] += 1
				details.append({
					'id': jd['jobid'],
					'state': state['state'],
					'signature': jd['signature']
				})
			out = {'name': self._display_name, 'summary': summary, 'details': details}
			return json.dumps({'success': out}, default=str)

	def __get_one_json(self, n):
		j = self.sched.get_job_by_id(n)
		if j is None:
			return json.dumps({'error':'Invalid job id'})
		return json.dumps({'success': j.to_dict()}, default=str)

	def __show_all(self):
		if len(self.sched.jobs)==0:
			return 'Nothing here'
		d = []
		table_id = 'all-jobs'
		for j in self.sched.jobs:
			jd = j.to_dict()
			duration = self.__duration(jd)
			state = self.__state(jd)
			start_dt = jd['logs']['start']
			end_dt = jd['logs']['end']
			next_dt = jd['next_run']
			next_dt_str = ''
			if next_dt is None:
				if jd['is_disabled']:
					next_dt_str = 'Disabled'
				else:
					next_dt_str = 'Never'
			else:
				next_dt_str = self.__date_fmt(next_dt)

			d.append(OrderedDict({
				'Id': TD(jd['jobid']),
				'Name': TD(jd['func'].replace('<', '&lt;').replace('>', '&gt;'), attrs={'title':j.func_signature()}),
				'Schedule': TD(self.__schedule_str(jd)),
				'Description': self.__descrTD(jd['doc']),
				'State': TD(state['state'], css=state['css'], attrs={'title': state['title']}),
				'Start': TD(self.__date_fmt(start_dt), attrs=self.__date_sort_attr(start_dt)),
				'End': TD(self.__date_fmt(end_dt), attrs=self.__date_sort_attr(end_dt)),
				'Time Taken': TD(duration),
				'Next Run': TD(next_dt_str, attrs=self.__date_sort_attr(next_dt)),
				'More':TD("<a href='/{}/{}'><button>show more</button><a>".format(self._endpoint, jd['jobid']))
			}))
		rows = [TR(row.values()) for row in d]
		head = [TH(th, default_sort=(th=="Next Run") ) for th in d[0].keys()]	# apply sorting to 'next run'
		all_jobs_table = TABLE(thead=THEAD(head), tbody=TBODY(rows), elem_id=table_id)
		rerun_txt = SMALL(f"Auto-refresh in {SPAN(self._homepage_refresh, attrs={'id': 'refresh-msg'})} seconds")
		filter_input = INPUT("", attrs={'type':'text', 'placeholder':'Filter', 'id': 'filter-box'})

		js_auto_reload_variables = '''
		new Tablesort(document.getElementById('{}'));
		let COUNT_DOWN = {};
		'''.format(table_id, self._homepage_refresh)

		js_functions = '''
		const table = document.getElementById("all-jobs")
		const filter_box = document.getElementById("filter-box");
		const tf = new TableFilter(table, filter_box)

		window.addEventListener('load', (event) => {
			const timer = setInterval(()=>{
				if (COUNT_DOWN > 0) {
					COUNT_DOWN --
					document.getElementById('refresh-msg').innerText = COUNT_DOWN
				} else {
					clearInterval(timer)
					location.reload()
				}
			}, 1000)
		})
		'''

		return self.__html_wrap(
			STYLES,
			H(2, "{} - Task Monitor".format(self._display_name)),
			SPAN("Running since {}".format(self._init_dt)),
			rerun_txt,
			filter_input,
			all_jobs_table,
			SCRIPT(js_auto_reload_variables + js_functions),
		)

	def __show_one(self, n):
		j = self.sched.get_job_by_id(n)
		if j is None:
			return 'Not found'
		jobd = j.to_dict()
		titleTD = lambda t: TD(t, css='title')
		state = self.__state(jobd)
		job_funcname = jobd['func'].replace('<', '&lt;').replace('>', '&gt;')

		enable_disable_btn = '''<button class="btn enable-disable-btn" onclick="enable_disable('{name}', {jobid}, {job_disable})" {btn_disabled}>{btn_name}</button>'''.format(
			name=job_funcname, jobid=n,
			job_disable="true" if not jobd['is_disabled'] else "false",
			btn_disabled="disabled" if state['state']=="RUNNING" else "",
			btn_name="Disable" if not jobd['is_disabled'] else "Enable",
		)
		rerun_btn = '''<button class="btn rerun-btn" onclick="rerun_trigger('{name}', {jobid})" {btn_disabled}>Rerun</button>'''.format(
			name=job_funcname, jobid=n, # rerun_trigger params
			btn_disabled="disabled" if state['state']=="RUNNING" or jobd['is_disabled'] else ""
		)
		rows = [
			TR([ titleTD("Schedule"), TD(self.__schedule_str(jobd)), ]),
			TR([ titleTD("State"), TD(state['state'], css=state['css']) ]),
			TR([ titleTD("Start Time"), TD(self.__date_fmt(jobd['logs']['start'])) ]),
			TR([ titleTD("End Time"), TD(self.__date_fmt(jobd['logs']['end'])) ]),
			TR([ titleTD("Time Taken"), TD(self.__duration(jobd)) ]),
			TR([ titleTD("Next Run In"), "<td id='next-run-in'>-<td>" ]),
			TR([ TD(enable_disable_btn, colspan=2, attrs={'style':'text-align:center'}) ]) if self._can_disable else '',
			TR([ TD(rerun_btn, colspan=2, attrs={'style':'text-align:center'}) ]) if self._can_rerun else ''
		]

		info_table = TABLE(tbody=TBODY(rows), css='info_table')
		description_div = DIV( CODE(jobd['src'], css='python'), css=['console-color ', 'console-div', 'brdr'])
		title = H(2, job_funcname, attrs={'title': j.func_signature()})
		monitor_div = DIV(
			title + info_table + description_div,
			css="monitor"
		)

		logs_row = TR([
			TD( DIV( CODE(jobd['logs']['log'], css='accesslog'), css='console-div'), css="console-color"),
			TD( DIV( CODE(jobd['logs']['err'], css='accesslog'), css='console-div'), css="console-color"),
		])
		logs_table = TABLE(thead=THEAD([TH('Logs'), TH('Traceback')]), tbody=TBODY(logs_row), css='log_table')
		logs_div = DIV( logs_table, css="logs_div" )

		container = DIV(
			monitor_div + logs_div,
			css="container"
		)

		next_run_ts = '"Never"'
		if jobd['is_disabled']:
			next_run_ts = '"Disabled"'
		elif jobd['next_run']:
			next_run_ts = jobd['next_run'].timestamp()

		auto_reload_script = '''
		Number.prototype.pad = function(size) {{
			let s = String(this);
			while (s.length < (size || 2)) {{s = "0" + s;}}
			return s;
		}}
		let running = {is_running}
		let next_run = {next_run_ts}
		let err_line = {err_line}
		function countdown_str(seconds) {{
			let hours = Math.floor(seconds / (60*60))
			seconds -= hours * (60*60)
			let minutes = Math.floor(seconds / 60)
			seconds -= minutes * 60
			return `${{hours.pad()}}:${{minutes.pad()}}:${{Math.floor(seconds).pad()}}`
		}}
		function rerun_trigger(job_name, jobid) {{
			const input_txt = prompt("Please type in the job name to confirm rerun", "");
			console.log(jobid)
			const payload = {{jobid, api_token:'{api_token}'}}
			if (input_txt===job_name) {{
				fetch('./rerun', {{method: 'POST', body: JSON.stringify(payload)}}).then(resp => {{
					return resp.json();
				}}).then(j=>{{
					if (j.success)
						window.location.reload()
					else if (j.error)
						throw Error(j.error)
					else
						throw Error("Rerun failed")
				}}).catch(e=>alert(e))
			}} else {{
				alert("Rerun aborted")
			}}
		}}
		function enable_disable(job_name, jobid, disable) {{
			const prompt_txt = "Please type in the job name to confirm " + ((disable) ? "disable": "enable")
			const input_txt = prompt(prompt_txt, "");
			console.log(jobid)
			const payload = {{jobid, disable, api_token:'{api_token}'}}
			if (input_txt===job_name) {{
				fetch('./enable_disable', {{method: 'POST', body: JSON.stringify(payload)}}).then(resp => {{
					return resp.json();
				}}).then(j=>{{
					if (j.success)
						window.location.reload()
					else if (j.error)
						throw Error(j.error)
					else
						throw Error("Action failed")
				}}).catch(e=>alert(e))
			}} else {{
				alert("Action aborted")
			}}
		}}
		window.addEventListener('load', (event) => {{
			//scroll to bottom
			document.getElementsByClassName("log_table")[0].querySelectorAll("div").forEach(d=>d.scrollTo(0,d.scrollHeight))
			if (running) {{
				setTimeout(()=>location.reload(), {taskpage_refresh}000)
			}} else if ( isNaN(next_run) ) {{ // if not number
				document.getElementById("next-run-in").innerHTML = next_run
			}} else {{
				const timer = setInterval(()=>{{
					let ttr = (next_run * 1000)-Date.now()
					if (ttr<=0) {{
						clearInterval(timer)
						setTimeout(()=>location.reload(), 1000) // small timeout to avoid too many reloads
					}} else {{
						document.getElementById("next-run-in").innerHTML = countdown_str(ttr/1000)
					}}
				}}, 1000)
			}}
			//highlight error line
			if (err_line>=0) {{
				hljs.initHighlightLinesOnLoad([
					[{{start: err_line, end: err_line, color: 'rgba(255, 0, 0, 0.4)'}}], // Highlight some lines in the first code block.
				]);
			}}
		}});
		'''.format(
			is_running=int(jobd['is_running']),
			next_run_ts=next_run_ts,
			err_line=self.__src_err_line(jobd),
			taskpage_refresh=self._taskpage_refresh,
			api_token=self._api_protection_token
		)

		return self.__html_wrap(
			STYLES,
			container,
			SCRIPT(auto_reload_script)
		)


	def __rerun_job(self):
		error = None
		data = json.loads(request.data)
		print("> rerun", data)
		if 'api_token' not in data or data['api_token']!=self._api_protection_token:
			error = 'Invalid token. Rerun blocked. Please reload the page and try again'

		elif 'jobid' not in data or not isinstance(data['jobid'], int):
			error = 'Invalid input'
		else:
			try:
				self.sched.rerun(data['jobid'])
			except Exception as e:
				error = str(e)

		if error is not None:
			return json.dumps({'error': error})
		else:
			return json.dumps({'success': True})


	def __enable_disable_job(self):
		error = None
		data = json.loads(request.data)
		print("> enable_disable", data)
		if 'api_token' not in data or data['api_token']!=self._api_protection_token:
			error = 'Action blocked'

		if 'jobid' not in data or not isinstance(data['jobid'], int):
			error = 'Invalid input'
		else:
			try:
				j = self.sched.get_job_by_id(data['jobid'])
				if j is None:
					raise ValueError("Job not found")
				if data['disable']==True:
					j.disable()
				else:
					j.enable()
			except Exception as e:
				error = str(e)

		if error is not None:
			return json.dumps({'error': error})
		else:
			return json.dumps({'success': True})



import holidays
from datetime import date
from dateutil.relativedelta import relativedelta, MO, FR
# from dateutil.easter import easter


class TradingHolidays(holidays.countries.UnitedStates):
	def _populate(self, year):
		# Populate the holiday list with the default US holidays
		holidays.UnitedStates._populate(self, year)
		# Remove Columbus Day
		self.pop(date(year, 10, 1) + relativedelta(weekday=MO(+2)), None)
		# Remove Veterans Day
		self.pop(date(year, 11, 11), None)
		if year==2023:
			self.pop(date(year, 11, 10), None)

		# # Add Good Friday # NOTE: removed this since Good Friday is not a bond market holiday - [04/01/2021]
		# self[easter(year) + relativedelta(weekday=FR(-1))] = "Good Friday"

		# 2021-12-31 is not a holiday apparently :( and windows version seems to think it is
		nye_2021 = date(2021, 12, 31)
		if nye_2021 in self:
			self.pop(nye_2021, None)




def HTML(content, title):
	return '''
	<!DOCTYPE html>
	<html lang="en">
		<head>
			<meta charset="UTF-8">
			<meta name="viewport" content="width=device-width, initial-scale=1.0">
			<script src="https://cdnjs.cloudflare.com/ajax/libs/tablesort/5.2.1/tablesort.min.js"
				integrity="sha512-F/gIMdDfda6OD2rnzt/Iyp2V9JLHlFQ+EUyixDg9+rkwjqgW1snpkpx7FD5FV1+gG2fmFj7I3r6ReQDUidHelA=="
				crossorigin="anonymous"></script>
			<script src="https://cdnjs.cloudflare.com/ajax/libs/tablesort/5.2.1/sorts/tablesort.number.min.js"
				integrity="sha512-dRD755QRxlybm0h3LXXIGrFcjNakuxW3reZqnPtUkMv6YsSWoJf+slPjY5v4lZvx2ss+wBZQFegepmA7a2W9eA=="
				crossorigin="anonymous"></script>
			<link rel="stylesheet"
				href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.6.0/styles/monokai-sublime.min.css">
			<script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.6.0/highlight.min.js"></script>
			<script src="https://cdn.jsdelivr.net/gh/TRSasasusu/highlightjs-highlight-lines.js@1.2.0/highlightjs-highlight-lines.min.js"></script>
			<script src="https://cdn.jsdelivr.net/gh/shashfrankenstien/lib-tablefilterjs@v0.0.5/lib-tablefilter.min.js"></script>
			<script>
				hljs.configure({{languages: ['python', 'accesslog']}});
				hljs.initHighlightingOnLoad();
			</script>
			<title>{}</title>
		</head>
		<body>
			{}
		</body>
	</html>'''.format(title, str(content))

def _TAG(tag, content, css, attrs):
	attrs = ["{}='{}'".format(k,v) for k,v in attrs.items()]
	if not isinstance(css, (list,set,tuple)):
		css = [css]
	attrs.append("class='{}'".format(' '.join(css)))
	return "<{t} {a}>{c}</{t}>".format(t=tag, a=' '.join(attrs), c=content)

def H(index, content, css=[], attrs={}):
	return _TAG('h{i}'.format(i=index), content, css, attrs)

def B(content, css=[], attrs={}):
	return _TAG('b', content, css, attrs)

def I(content, css=[], attrs={}):
	return _TAG('i', content, css, attrs)

def SMALL(content, css=[], attrs={}):
	return _TAG('small', content, css, attrs)

def SPAN(content, css=[], attrs={}):
	return _TAG('span', content, css, attrs)

def DIV(content, css=[], attrs={}):
	return _TAG('div', content, css, attrs)

def TABLE(thead='', tbody='', elem_id='', css=[], attrs={}):
	attrs['id'] = elem_id
	return _TAG('table', '{}{}'.format(thead, tbody), css, attrs)

def TH(h, default_sort=False):
	return "<th {}>{}</th>".format("data-sort-default" if default_sort else "", h)

def THEAD(th, css=[], attrs={}):
	return _TAG('thead', ''.join(th), css, attrs)

def TBODY(rows, css=[], attrs={}):
	return _TAG('tbody', ''.join(rows), css, attrs)

def TD(content, colspan=1, rowspan=1, css=[], attrs={}):
	attrs['colspan'] = colspan
	attrs['rowspan'] = rowspan
	return _TAG('td', content if content is not None else "-", css, attrs)

def TR(row, css=[], attrs={}):
	return _TAG('tr', ''.join(row), css, attrs)

def INPUT(content, css=[], attrs={}):
	return _TAG('input', content, css, attrs)

def SCRIPT(s):
	return "<script>{}</script>".format(s)

def CODE(s, css=[]):
	if not isinstance(css, (list,set,tuple)):
		css = [css]
	return "<pre><code class='{}'>{}</code></pre>".format(' '.join(css), s)
